from django.conf import settings


EXPOSE_MODEL_TO_ADMIN = getattr(settings, 'SITEPREFS_EXPOSE_MODEL_TO_ADMIN', False)
"""Toggles internal preferences model showing up in the Admin."""

DISABLE_AUTODISCOVER = getattr(settings, 'SITEPREFS_DISABLE_AUTODISCOVER', False)
"""Disables preferences autodiscovery on Django apps registry ready."""

PREFS_MODULE_NAME = getattr(settings, 'SITEPREFS_MODULE_NAME', 'settings')
"""Module name used by siteprefs.toolbox.autodiscover_siteprefs() to find preferences in application packages."""

MANAGE_SAFE_COMMANDS = getattr(
    settings, 'SITEPREFS_MANAGE_SAFE_COMMANDS', ['runserver', 'runserver_plus', 'run_gunicorn', 'celeryd'])
"""List of commands called by manage.py where autodiscover is enabled"""



import sys
from collections import OrderedDict
from types import ModuleType
from typing import Dict, Union, List, Tuple, Any, Optional

from django.contrib import admin
from django.contrib.admin import AdminSite
from django.db.models import Model, Field

from .exceptions import SitePrefsException
from .models import Preference
from .settings import MANAGE_SAFE_COMMANDS
from .signals import prefs_save
from .utils import import_prefs, get_frame_locals, traverse_local_prefs, get_pref_model_admin_class, \
    get_pref_model_class, PrefProxy, PatchedLocal, Frame

__PATCHED_LOCALS_SENTINEL = '__siteprefs_locals_patched'

__PREFS_REGISTRY = None
__PREFS_DEFAULT_REGISTRY = OrderedDict()
__MODELS_REGISTRY = {}


def on_pref_update(*args, **kwargs):
    """Triggered on dynamic preferences model save.
     Issues DB save and reread.

    """
    Preference.update_prefs(*args, **kwargs)
    Preference.read_prefs(get_prefs())

prefs_save.connect(on_pref_update)


def get_prefs() -> dict:
    """Returns a dictionary with all preferences discovered by siteprefs."""
    global __PREFS_REGISTRY

    if __PREFS_REGISTRY is None:
        __PREFS_REGISTRY = __PREFS_DEFAULT_REGISTRY

    return __PREFS_REGISTRY


def get_app_prefs(app: str = None) -> dict:
    """Returns a dictionary with preferences for a certain app/module.

    :param app:

    """
    if app is None:

        with Frame(stepback=1) as frame:
            app = frame.f_globals['__name__'].split('.')[0]

    prefs = get_prefs()

    if app not in prefs:
        return {}

    return prefs[app]


def get_prefs_models() -> Dict[str, Model]:
    """Returns registered preferences models indexed by application names."""
    return __MODELS_REGISTRY


def bind_proxy(
        values: Union[List, Tuple],
        category: str = None,
        field: Field = None,
        verbose_name: str = None,
        help_text: str = '',
        static: bool = True,
        readonly: bool = False
) -> List[PrefProxy]:
    """Binds PrefProxy objects to module variables used by apps as preferences.

    :param values: Preference values.

    :param category: Category name the preference belongs to.

    :param field: Django model field to represent this preference.

    :param verbose_name: Field verbose name.

    :param help_text: Field help text.

    :param static: Leave this preference static (do not store in DB).

    :param readonly: Make this field read only.

    """
    addrs = OrderedDict()

    depth = 3

    for local_name, locals_dict in traverse_local_prefs(depth):
        addrs[id(locals_dict[local_name])] = local_name

    proxies = []
    locals_dict = get_frame_locals(depth)

    for value in values:  # Try to preserve fields order.

        id_val = id(value)

        if id_val in addrs:
            local_name = addrs[id_val]
            local_val = locals_dict[local_name]

            if isinstance(local_val, PatchedLocal) and not isinstance(local_val, PrefProxy):

                proxy = PrefProxy(
                    local_name, value.val,
                    category=category,
                    field=field,
                    verbose_name=verbose_name,
                    help_text=help_text,
                    static=static,
                    readonly=readonly,
                )

                app_name = locals_dict['__name__'].split('.')[-2]  # x.y.settings -> y
                prefs = get_prefs()

                if app_name not in prefs:
                    prefs[app_name] = OrderedDict()

                prefs[app_name][local_name.lower()] = proxy

                # Replace original pref variable with a proxy.
                locals_dict[local_name] = proxy
                proxies.append(proxy)

    return proxies


def register_admin_models(admin_site: AdminSite):
    """Registers dynamically created preferences models for Admin interface.

    :param admin_site: AdminSite object.

    """
    global __MODELS_REGISTRY

    prefs = get_prefs()

    for app_label, prefs_items in prefs.items():

        model_class = get_pref_model_class(app_label, prefs_items, get_app_prefs)

        if model_class is not None:
            __MODELS_REGISTRY[app_label] = model_class
            admin_site.register(model_class, get_pref_model_admin_class(prefs_items))


def autodiscover_siteprefs(admin_site: AdminSite = None):
    """Automatically discovers and registers all preferences available in all apps.

    :param admin_site: Custom AdminSite object.

    """
    if admin_site is None:
        admin_site = admin.site

    # Do not discover anything if called from manage.py (e.g. executing commands from cli).
    if 'manage' not in sys.argv[0] or (len(sys.argv) > 1 and sys.argv[1] in MANAGE_SAFE_COMMANDS):
        import_prefs()
        Preference.read_prefs(get_prefs())
        register_admin_models(admin_site)


def patch_locals(depth: int = 2):
    """Temporarily (see unpatch_locals()) replaces all module variables
    considered preferences with PatchedLocal objects, so that every
    variable has different hash returned by id().

    """
    for name, locals_dict in traverse_local_prefs(depth):
        locals_dict[name] = PatchedLocal(name, locals_dict[name])

    get_frame_locals(depth)[__PATCHED_LOCALS_SENTINEL] = True  # Sentinel.


def unpatch_locals(depth: int = 3):
    """Restores the original values of module variables
    considered preferences if they are still PatchedLocal
    and not PrefProxy.

    """
    for name, locals_dict in traverse_local_prefs(depth):
        if isinstance(locals_dict[name], PatchedLocal):
            locals_dict[name] = locals_dict[name].val

    del get_frame_locals(depth)[__PATCHED_LOCALS_SENTINEL]


class ModuleProxy:
    """Proxy to handle module attributes access."""

    def __init__(self):
        self._module: Optional[ModuleType] = None
        self._prefs = []

    def bind(self, module: ModuleType, prefs: List[str]):
        """
        :param module:
        :param prefs: Preference names. Just to speed up __getattr__.

        """
        self._module = module
        self._prefs = set(prefs)

    def __getattr__(self, name: str) -> Any:

        value = getattr(self._module, name)

        if name in self._prefs:
            # It is a PrefProxy
            value = value.value

        return value


def proxy_settings_module(depth: int = 3):
    """Replaces a settings module with a Module proxy to intercept
    an access to settings.

    :param depth: Frame count to go backward.

    """
    proxies = []

    modules = sys.modules
    module_name = get_frame_locals(depth)['__name__']

    module_real = modules[module_name]

    for name, locals_dict in traverse_local_prefs(depth):

        value = locals_dict[name]

        if isinstance(value, PrefProxy):
            proxies.append(name)

    new_module = type(module_name, (ModuleType, ModuleProxy), {})(module_name)  # ModuleProxy
    new_module.bind(module_real, proxies)

    modules[module_name] = new_module


def register_prefs(*args: PrefProxy, **kwargs):
    """Registers preferences that should be handled by siteprefs.

    Expects preferences as *args.

    Use keyword arguments to batch apply params supported by
    ``PrefProxy`` to all preferences not constructed by ``pref`` and ``pref_group``.

    Batch kwargs:

        :param str help_text: Field help text.

        :param bool static: Leave this preference static (do not store in DB).

        :param bool readonly: Make this field read only.

    :param bool swap_settings_module: Whether to automatically replace settings module
        with a special ``ProxyModule`` object to access dynamic values of settings
        transparently (so not to bother with calling ``.value`` of ``PrefProxy`` object).

    """
    swap_settings_module = bool(kwargs.get('swap_settings_module', True))

    if __PATCHED_LOCALS_SENTINEL not in get_frame_locals(2):
        raise SitePrefsException('Please call `patch_locals()` right before the `register_prefs()`.')

    bind_proxy(args, **kwargs)

    unpatch_locals()

    swap_settings_module and proxy_settings_module()


def pref_group(
        title: str,
        prefs: Union[List, Tuple],
        help_text: str = '',
        static: bool = True,
        readonly: bool = False
):
    """Marks preferences group.

    :param title: Group title

    :param prefs: Preferences to group.

    :param help_text: Field help text.

    :param static: Leave this preference static (do not store in DB).

    :param readonly: Make this field read only.

    """
    bind_proxy(prefs, title, help_text=help_text, static=static, readonly=readonly)

    for proxy in prefs:  # For preferences already marked by pref().
        if isinstance(proxy, PrefProxy):
            proxy.category = title


def pref(
        preference: Any,
        field: Field = None,
        verbose_name: str = None,
        help_text: str = '',
        static: bool = True,
        readonly: bool = False
) -> Optional[PrefProxy]:
    """Marks a preference.

    :param preference: Preference variable.

    :param field: Django model field to represent this preference.

    :param verbose_name: Field verbose name.

    :param help_text: Field help text.

    :param static: Leave this preference static (do not store in DB).

    :param readonly: Make this field read only.

    """
    try:
        bound = bind_proxy(
            (preference,),
            field=field,
            verbose_name=verbose_name,
            help_text=help_text,
            static=static,
            readonly=readonly,
        )
        return bound[0]

    except IndexError:
        return


class preferences:
    """Context manager - main entry point for siteprefs.

    .. code-block:: python

        from siteprefs.toolbox import preferences

        with preferences() as prefs:

            prefs(
                MY_OPTION_1,
                prefs.one(MY_OPTION_2, static=False),
                prefs.group('My Group', [prefs.one(MY_OPTION_42)]),
            )

    """

    one = staticmethod(pref)
    group = staticmethod(pref_group)

    __call__ = register_prefs

    def __enter__(self):
        patch_locals(3)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass




import inspect
import os
from collections import OrderedDict
from datetime import datetime
from typing import Any, Callable, Type, Generator, Tuple
from warnings import warn

from django.contrib import admin
from django.db import models
from django.utils.translation import gettext_lazy as _
from etc.toolbox import import_app_module, import_project_modules

from .settings import PREFS_MODULE_NAME
from .signals import prefs_save


class Frame:
    """Represents a frame object at a definite level of hierarchy.

    To be used as context manager:

        with Frame as f:
            ...

    """

    def __init__(self, stepback: int = 0):
        self.depth = stepback

    def __enter__(self):
        frame = inspect.currentframe().f_back

        for __ in range(self.depth):
            frame = frame.f_back

        self.frame = frame

        return self.frame

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.frame


class PatchedLocal:
    """Object of this class temporarily replace all module variables
    considered preferences.

    """
    def __init__(self, key: str, val: Any):
        self.key = key
        self.val = val


class Mimic:
    """Mimics other types by implementation of various special methods.

    This one is deprecated if favor of setting module proxying (proxy_settings_module()).

    """

    value: Any = None

    def __call__(self, *args, **kwargs):
        return self.value

    def __str__(self):
        return self.value.__str__()

    def __bool__(self):
        return bool(self.value)

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __len__(self):
        return self.value.__len__()

    def __contains__(self, item):
        return self.value.__contains__(item)

    def __sub__(self, other):
        return self.value.__sub__(other)

    def __rsub__(self, other):
        return self.value.__rsub__(other)

    def __add__(self, other):
        return self.value.__add__(other)

    def __radd__(self, other):
        return self.value.__radd__(other)

    def __mul__(self, other):
        return self.value.__mul__(other)

    def __rmul__(self, other):
        return self.value.__rmul__(other)

    def __lt__(self, other):
        return self.value.__lt__(other)

    def __le__(self, other):
        return self.value.__le__(other)

    def __gt__(self, other):
        return self.value.__gt__(other)

    def __ge__(self, other):
        return self.value.__ge__(other)

    def __eq__(self, other):
        return self.value.__eq__(other)

    def __ne__(self, other):
        return self.value.__ne__(other)


class PrefProxy(Mimic):
    """Objects of this class replace app preferences."""

    def __init__(
            self,
            name: str,
            default: Any,
            category: str = None,
            field: models.Field = None,
            verbose_name: str = None,
            help_text: str = '',
            static: bool = True,
            readonly: bool = False
    ):
        """

        :param name: Preference name.

        :param default: Default (initial) value.

        :param category: Category name the preference belongs to.

        :param field: Django model field to represent this preference.

        :param verbose_name: Field verbose name.

        :param help_text: Field help text.

        :param static: Leave this preference static (do not store in DB).

        :param readonly: Make this field read only.

        """
        self.name = name
        self.category = category
        self.default = default
        self.static = static
        self.help_text = help_text

        if static:
            readonly = True

        self.readonly = readonly

        if verbose_name is None:
            verbose_name = name.replace('_', ' ').capitalize()

        self.verbose_name = verbose_name

        if field is None:
            self.field = get_field_for_proxy(self)

        else:
            self.field = field
            update_field_from_proxy(self.field, self)

    @property
    def value(self) -> Any:

        if self.static:
            val = self.default

        else:
            try:
                val = getattr(self, 'db_value')

            except AttributeError:
                val = self.default

        return self.field.to_python(val)

    def get_value(self) -> Any:
        warn('Please use .value instead .get_value().', DeprecationWarning, stacklevel=2)
        return self.value

    def __repr__(self):
        return f'{self.name} = {self.value}'


def get_field_for_proxy(pref_proxy: PrefProxy) -> models.Field:
    """Returns a field object instance for a given PrefProxy object.

    :param pref_proxy:

    """
    field = {
        bool: models.BooleanField,
        int:  models.IntegerField,
        float: models.FloatField,
        datetime: models.DateTimeField,

    }.get(type(pref_proxy.default), models.TextField)()

    update_field_from_proxy(field, pref_proxy)

    return field


def update_field_from_proxy(field_obj: models.Field, pref_proxy: PrefProxy):
    """Updates field object with data from a PrefProxy object.

    :param field_obj:

    :param pref_proxy:

    """
    attr_names = ('verbose_name', 'help_text', 'default')

    for attr_name in attr_names:
        setattr(field_obj, attr_name, getattr(pref_proxy, attr_name))


def get_pref_model_class(app: str, prefs: dict, get_prefs_func: Callable) -> Type[models.Model]:
    """Returns preferences model class dynamically crated for a given app or None on conflict."""

    module = f'{app}.{PREFS_MODULE_NAME}'

    model_dict = {
            '_prefs_app': app,
            '_get_prefs': staticmethod(get_prefs_func),
            '__module__': module,
            'Meta': type('Meta', (models.options.Options,), {
                'verbose_name': _('Preference'),
                'verbose_name_plural': _('Preferences'),
                'app_label': app,
                'managed': False,
            })
    }

    for field_name, val_proxy in prefs.items():
        model_dict[field_name] = val_proxy.field

    model = type('Preferences', (models.Model,), model_dict)

    def fake_save_base(self, *args, **kwargs):

        updated_prefs = {
            f.name: getattr(self, f.name)
            for f in self._meta.fields
            if not isinstance(f, models.fields.AutoField)
        }

        app_prefs = self._get_prefs(self._prefs_app)

        for pref in app_prefs.keys():
            if pref in updated_prefs:
                app_prefs[pref].db_value = updated_prefs[pref]

        self.pk = self._prefs_app  # Make Django 1.7 happy.
        prefs_save.send(sender=self, app=self._prefs_app, updated_prefs=updated_prefs)

        return True

    model.save_base = fake_save_base

    return model


def get_pref_model_admin_class(prefs: dict) -> Type[admin.ModelAdmin]:

    by_category = OrderedDict()
    readonly_fields = []

    for field_name, val_proxy in prefs.items():

        if val_proxy.readonly:
            readonly_fields.append(field_name)

        if val_proxy.category not in by_category:
            by_category[val_proxy.category] = []

        by_category[val_proxy.category].append(field_name)

    cl_model_admin_dict = {
        'has_add_permission': lambda *args: False,
        'has_delete_permission': lambda *args: False
    }

    if readonly_fields:
        cl_model_admin_dict['readonly_fields'] = readonly_fields

    fieldsets = []

    for category, cat_prefs in by_category.items():
        fieldsets.append((category, {'fields': cat_prefs}))

    if fieldsets:
        cl_model_admin_dict['fieldsets'] = fieldsets

    model = type('PreferencesAdmin', (admin.ModelAdmin,), cl_model_admin_dict)

    model.changelist_view = lambda self, request, **kwargs: self.change_view(request, '', **kwargs)

    model.get_object = lambda self, *args: (
        self.model(
            **{
                field_name: val_proxy.get_value() for field_name, val_proxy in
                self.model._get_prefs(self.model._prefs_app).items()
            }
        )
    )

    return model


def get_frame_locals(stepback: int = 0) -> dict:
    """Returns locals dictionary from a given frame.

    :param stepback:

    """
    with Frame(stepback=stepback) as frame:
        locals_dict = frame.f_locals

    return locals_dict


def traverse_local_prefs(stepback: int = 0) -> Generator[Tuple[str, dict], None, None]:
    """Generator to walk through variables considered as preferences
    in locals dict of a given frame.

    :param stepback:

    """
    locals_dict = get_frame_locals(stepback+1)
    for k in locals_dict:
        if not k.startswith('_') and k.upper() == k:
            yield k, locals_dict


def import_module(package: str, module_name: str):
    """Imports a module from a given package.

    :param package:
    :param module_name:

    """
    import_app_module(package, module_name)


def import_prefs():
    """Imports preferences modules from packages (apps) and project root."""
    
    # settings.py locals if autodiscover_siteprefs() is in urls.py
    settings_locals = get_frame_locals(3)

    if 'self' not in settings_locals:  # If not SiteprefsConfig.ready()
        # Try to import project-wide prefs.

        project_package = settings_locals['__package__']  # Expected project layout introduced in Django 1.4
        if not project_package:
            # Fallback to old layout.
            project_package = os.path.split(os.path.dirname(settings_locals['__file__']))[-1]

        import_module(project_package, PREFS_MODULE_NAME)

    import_project_modules(PREFS_MODULE_NAME)



from django.db import models
from django.db.utils import IntegrityError
from django.utils.translation import gettext_lazy as _


class Preference(models.Model):

    app = models.CharField(_('Application'), max_length=100, null=True, blank=True, db_index=True)
    name = models.CharField(_('Name'), max_length=150)
    text = models.TextField(_('Value'), null=True, blank=True)

    class Meta:
        verbose_name = _('Preference')
        verbose_name_plural = _('Preferences')
        unique_together = ('app', 'name')

    def __str__(self):
        return f'{self.app}:{self.name}'

    @classmethod
    def read_prefs(cls, mem_prefs: dict):
        """Initializes preferences entries in DB according to currently discovered prefs.

        :param mem_prefs:

        """
        db_prefs = {
            f"{pref['app']}__{pref['name']}": pref for pref in
            cls.objects.values().order_by('app', 'name')
        }

        new_prefs = []

        for app, prefs in mem_prefs.items():

            for pref_name, pref_proxy in prefs.items():

                if not pref_proxy.static:  # Do not add static options to DB.
                    key = f'{app}__{pref_name}'

                    if key in db_prefs:
                        # Entry already exists in DB. Let's get pref value from there.
                        pref_proxy.db_value = db_prefs[key]['text']

                    else:
                        new_prefs.append(cls(app=app, name=pref_name, text=pref_proxy.default))

        if new_prefs:
            try:
                cls.objects.bulk_create(new_prefs)

            except IntegrityError:  # Don't bother with duplicates.
                pass

    @classmethod
    def update_prefs(cls, *args, **kwargs):
        # TODO That could be more efficient.

        updated_prefs = kwargs['updated_prefs']

        for db_pref in cls.objects.filter(app=kwargs['app']):

            if db_pref.name in updated_prefs:
                db_pref.text = updated_prefs[db_pref.name]
                db_pref.save()



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

from .settings import DISABLE_AUTODISCOVER


class SiteprefsConfig(AppConfig):
    """The default siteprefs configuration."""

    name = 'siteprefs'
    verbose_name = _('Site Preferences')

    def ready(self):

        if DISABLE_AUTODISCOVER:
            return

        from .toolbox import autodiscover_siteprefs
        autodiscover_siteprefs()



import pytest
from django.contrib import admin
from django.db import models

from siteprefs.models import Preference
from siteprefs.toolbox import autodiscover_siteprefs, get_app_prefs, get_prefs_models
from siteprefs.utils import Frame, PatchedLocal, PrefProxy, get_field_for_proxy, get_pref_model_class, \
    get_pref_model_admin_class, get_frame_locals, import_module, PREFS_MODULE_NAME


def test_many():
    __package__ = 'siteprefs.tests.testapp'

    autodiscover_siteprefs()

    prefs = list(Preference.objects.all())
    assert len(prefs) == 1

    pref = prefs[0]
    assert pref.name == 'my_option_2'
    assert pref.text == 'Some value'

    prefs = get_app_prefs('testapp')
    assert len(prefs) == 3

    prefs = get_app_prefs('buggy')
    assert len(prefs) == 0

    def update_option_2(new_value):
        # Test update prefs triggering.
        model_cls = get_prefs_models()['testapp']
        model = model_cls()
        model.my_option_2 = new_value
        model.save()

    update_option_2('other_value')

    pref = list(Preference.objects.all())[0]
    assert pref.text == 'other_value'
    assert '%s' % pref == 'testapp:my_option_2'

    from siteprefs.tests.testapp import testmodule

    assert testmodule.read_option_2() == 'other_value'
    assert testmodule.read_not_an_option() == 'not-an-option'

    update_option_2(44)

    assert testmodule.read_option_2() == '44'


def test_admin():
    from siteprefs.admin import PreferenceAdmin
    return PreferenceAdmin  # not too loose unused import


class TestUtils:

    def test_frame(self):
        with Frame() as f:
            assert f.f_locals['self'] is self

    def test_patched_local(self):
        pl = PatchedLocal('k', 'v')
        assert pl.key == 'k'
        assert pl.val == 'v'

    def test_pref_poxy(self):
        pp = PrefProxy('proxy_name', 'default')
        assert pp.name == 'proxy_name'
        assert pp.default == 'default'
        assert pp.category is None
        assert isinstance(pp.field, models.TextField)
        assert pp.verbose_name == 'Proxy name'
        assert pp.help_text == ''
        assert pp.static
        assert pp.readonly
        assert pp.get_value() == 'default'

        pp = PrefProxy(
            'proxy_name_2', 2,
            category='cat',
            field=models.IntegerField(),
            verbose_name='verbose name',
            help_text='help text',
            static=False,
            readonly=True)

        assert pp.name == 'proxy_name_2'
        assert pp.default == 2
        assert pp.category == 'cat'
        assert isinstance(pp.field, models.IntegerField)
        assert pp.verbose_name == 'verbose name'
        assert pp.help_text == 'help text'
        assert not pp.static
        assert pp.readonly
        assert pp.get_value() == 2

        pp.db_value = 42
        assert pp.get_value() == 42
        assert pp() == 42
        assert pp() < 43
        assert pp() > 41

        pp = PrefProxy('proxy_name_3', 10, static=False)
        assert not pp.readonly

    def test_get_field_for_proxy(self):
        pp = PrefProxy('proxy_name', 10)
        assert isinstance(get_field_for_proxy(pp), models.IntegerField)

        pp = PrefProxy('proxy_name', True)
        assert isinstance(get_field_for_proxy(pp), models.BooleanField)

        pp = PrefProxy('proxy_name', 13.4)
        assert isinstance(get_field_for_proxy(pp), models.FloatField)

        pp = PrefProxy('proxy_name', 'abc')
        assert isinstance(get_field_for_proxy(pp), models.TextField)

    def test_get_pref_model_class(self):
        p1 = PrefProxy('pp1', 10)
        p2 = PrefProxy('pp2', 20)
        p3 = PrefProxy('pp3', 'admin', field=models.CharField(max_length=10))
        p4 = PrefProxy('pp4', 'another', verbose_name='verbosed', static=False, field=models.CharField(max_length=10))

        my_prefs_func = lambda: 'yes'

        cl = get_pref_model_class('testapp', {'pp1': p1, 'pp2': p2, 'pp3': p3, 'pp4': p4}, my_prefs_func)

        model = cl()

        assert issubclass(cl, models.Model)
        assert isinstance(model._meta.fields[3], models.CharField)
        assert model._meta.fields[4].verbose_name == 'verbosed'

    def test_get_pref_model_admin_class(self):

        p1 = PrefProxy('pp1', 10)
        p2 = PrefProxy('pp2', 20)

        cl = get_pref_model_admin_class({'pp1': p1, 'pp2': p2})
        assert issubclass(cl, admin.ModelAdmin)

    def test_get_frame_locals(self):
        a = 'a'
        b = 'b'
        local = get_frame_locals(1)
        assert 'a' in local
        assert 'b' in local

    @pytest.mark.xfail(strict=True)
    def test_import_module(self):
        import_module('dummy', PREFS_MODULE_NAME)



from pytest_djangoapp import configure_djangoapp_plugin

pytest_plugins = configure_djangoapp_plugin({
    'SITEPREFS_EXPOSE_MODEL_TO_ADMIN': True,
    'SITEPREFS_DISABLE_AUTODISCOVER': True,
}, admin_contrib=True)



from django.conf import settings


NOT_AN_OPTION = 'not-an-option'
MY_OPTION_1 = getattr(settings, 'MY_APP_MY_OPTION_1', True)
MY_OPTION_2 = getattr(settings, 'MY_APP_MY_OPTION_2', 'Some value')
MY_OPTION_42 = getattr(settings, 'MY_APP_MY_OPTION_42', 42)


if 'siteprefs' in settings.INSTALLED_APPS:

    from siteprefs.toolbox import preferences

    with preferences() as prefs:

        prefs(
            MY_OPTION_1,
            prefs.one(MY_OPTION_2, static=False),
            prefs.group('My Group', [prefs.one(MY_OPTION_42)]),
        )



from siteprefs.utils import Mimic


def test_mimic():

    class M(Mimic):

        def __init__(self, value):
            self._val = value

        @property
        def value(self):
            return self._val

    bool(M(False))

    assert M(4)() == 4
    assert str(M('www')) == 'www'
    assert len(M('some')) == 4
    assert int(M(4)) == 4
    assert float(M(4.2)) == 4.2
    assert not bool(M(False))
    assert not M(False)
    assert M(True)
    assert (M(4) + 5) == 9
    assert (5 + M(4)) == 9
    assert (M(4) - 5) == -1
    assert (5 - M(4)) == 1
    assert (M(4) * 5) == 20
    assert (5 * M(4)) == 20

    x = 5
    x += M(4)
    assert x == 9

    x = M(4)
    x -= 1
    assert x == 3

    assert ('some%s' % M('any')) == 'someany'
    assert (M('any') + 'some') == 'anysome'
    assert M(4) < 5
    assert M(4) <= 5
    assert M(6) > 5
    assert M(6) >= 5
    assert M(5) == 5
    assert M(4) != 5
    assert len(M('some')) == 4
    assert 'ome' in M('some')
